'CLASSI PER L'ANIMAZIONE
Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

#Region "Animazione"
Public Class animazionClass
    Public Shared texPath As String

    Public currentTime As Single
    Public currentTrack As Byte
    Public animationSet() As AnimationSet
    Protected blendTime As Single
    'Classi di supporto
#Region "Allocated hierarchy"

    'Questa classe sovrascrive la classe di caricamento generico (astratta)
    'alla classe vengono passate da directX i dati sulle mesh che la classe intercetta
    'e mantiene in memoria
    Protected Class AH
        Inherits AllocateHierarchy

        Public Overrides Function CreateFrame(ByVal name As String) As Microsoft.DirectX.Direct3D.Frame
            'nuovo frame derivato
            Dim frame As frameD = New frameD
            frame.Name = name 'carico il nome
            frame.TransformationMatrix = Matrix.Identity 'matrici iniziali 
            frame.matrixC = Matrix.Identity 'matrici iniziali
            Return frame 'restituisce a directX il frame
        End Function

        Public Overrides Function CreateMeshContainer(ByVal name As String, ByVal meshData As Microsoft.DirectX.Direct3D.MeshData, ByVal materials() As Microsoft.DirectX.Direct3D.ExtendedMaterial, ByVal effectInstances() As Microsoft.DirectX.Direct3D.EffectInstance, ByVal adjacency As Microsoft.DirectX.GraphicsStream, ByVal skinInfo As Microsoft.DirectX.Direct3D.SkinInformation) As Microsoft.DirectX.Direct3D.MeshContainer
            If meshData.Mesh Is Nothing Then
                Err.Raise(777) 'se la mesh non contiene dati genera un errore
            End If

            'anche se i vertici sono vuoti genera
            If (meshData.Mesh.VertexFormat = VertexFormats.None) Then
                Err.Raise(777)
            End If

            'nuovo contenitore per la mesh
            Dim mesh As MeshContainerDerived = New MeshContainerDerived
            'assegna il nome al contenitore di mesh
            mesh.Name = name


            'correggi il formato
            'se non ci sono normali vengono generate
            If meshData.Mesh.VertexFormat And VertexFormats.Normal = False Then
                Dim tempM As Mesh = meshData.Mesh.Clone(meshData.Mesh.Options.Value, meshData.Mesh.VertexFormat Or VertexFormats.Normal, device)
                meshData.Mesh = tempM
                meshData.Mesh.ComputeNormals()
            End If

            'impostazione dati utili
            mesh.SetMaterials(materials) 'materiali
            mesh.SetAdjacency(adjacency) 'adiacenze

            'array di texture
            Dim meshTex() As Texture
            ReDim meshTex(materials.Length)

            'caricamento texture sistemare
            Dim i As Integer
            For i = 0 To materials.Length - 1
                If materials(i).TextureFilename <> "" Then
                    Try
                        'cerca di caricare la texture
                        meshTex(i) = TextureLoader.FromFile(device, texPath & "\" & materials(i).TextureFilename)
                    Catch ex As Exception
                        'la texture non viene caricata
                    End Try
                End If
            Next
            'assegno le texture
            mesh.meshTextures = meshTex

            'salvataggio dati
            mesh.MeshData = meshData

            ' se ci sono dati per lo skinning usali
            If Not (skinInfo Is Nothing) Then
                mesh.SkinInformation = skinInfo
                Dim numBones As Integer = skinInfo.NumberBones 'numero ossature
                Dim offsetMatrix() As Matrix 'matrici offset
                ReDim offsetMatrix(numBones)
                For i = 0 To numBones - 1
                    offsetMatrix(i) = skinInfo.GetBoneOffsetMatrix(i) 'le riceviamo direttamente dalle skinInfo
                Next
                mesh.offsetMatrices = offsetMatrix 'salvataggio matrice
                'converte la mesh in blended mesh
                animazionClass.GenerateSkinnedMesh(mesh)
            End If

            'restituisce la mesh

            Return mesh
        End Function


    End Class
#End Region

#Region "frame"
    'classe derivata da frame che contiene la matrice combinata
    'ossia la generazione della matrice in animazione
    Protected Class frameD
        Inherits Frame
        Public matrixC As Matrix 'matrice combinata

    End Class
#End Region

#Region "mesh container"
    'la classe deriv dal contenitore della mesh aggiungedovi alcuni dati
    Protected Class MeshContainerDerived
        Inherits MeshContainer

        Public meshTextures() As Texture 'texture
        Public numAttr As Integer = 0 'numero attributi
        Public numInfl As Integer = 0 'numero influenze delle bones
        Public bones() As BoneCombination 'bones
        Public frameMatrices() As frameD 'frame figli
        Public offsetMatrices() As Matrix 'matrici delle bones

    End Class
#End Region


    Public animazione As AnimationRootFrame 'contenitore della animazione
    Sub New(ByVal filesrc As String, ByVal texP As String)
        'settaggio
        texPath = texP
        Dim alloc As New AH 'classe per l'allocazione
        'caricamento animazione
        animazione = Mesh.LoadHierarchyFromFile(filesrc, MeshFlags.Managed, device, alloc, Nothing)
        'routine settaggio bone
        SetupBoneMatrices(CType(animazione.FrameHierarchy, frameD))
        storeAnimationSets()

    End Sub


    Private Overloads Sub SetupBoneMatrices(ByRef frame As frameD)
        'setta le matrici delle bones
        If Not (frame.MeshContainer Is Nothing) Then
            SetupBoneMatrices(CType(frame.MeshContainer, MeshContainerDerived))
        End If
        'se ci sono altre mesh nel contenitore setta le matrici delle bones
        If Not (frame.FrameSibling Is Nothing) Then
            SetupBoneMatrices(CType(frame.FrameSibling, frameD))
        End If
        'se ci sono mesh figlie setta le matrici
        If Not (frame.FrameFirstChild Is Nothing) Then
            SetupBoneMatrices(CType(frame.FrameFirstChild, frameD))
        End If

    End Sub



    Private Overloads Sub SetupBoneMatrices(ByRef mesh As MeshContainerDerived)
        ' se esiste l'oggetto skin information carica i dati delle bones
        If Not (mesh.SkinInformation Is Nothing) Then
            'numero bones
            Dim numBones As Integer = mesh.SkinInformation.NumberBones
            'array di frame figli
            Dim frameMatrices() As frameD
            ReDim frameMatrices(numBones)
            '
            Dim i As Integer
            For i = 0 To numBones - 1
                'per ogni bone trova il frame corrispondente
                Dim frame As frameD = frame.Find(animazione.FrameHierarchy, mesh.SkinInformation.GetBoneName(i))
                If frame Is Nothing Then
                    Return
                End If
                'salva 
                frameMatrices(i) = frame

            Next
            'salva i dati
            mesh.frameMatrices = frameMatrices
        End If
    End Sub

    Public Sub ProcessNextFrame(ByVal t As Single, ByVal matrice As Matrix)
        'elabora il successivo frame di animazione
        device.Transform.World = matrice
        'se esistente avanza il tempo
        animazione.AnimationController.AdvanceTime(t, Nothing)
        'aggiorna le matrici
        UpdateFrameMatrices(animazione.FrameHierarchy, matrice)
        currentTime += t
    End Sub




    Private Sub UpdateFrameMatrices(ByRef frame As frameD, ByVal parentMatrix As Matrix)
        'calcola la matrice combinata
        frame.matrixC = Matrix.Multiply(frame.TransformationMatrix, parentMatrix)
        'se ci sono altri frame nel contenitore
        If Not (frame.FrameSibling Is Nothing) Then
            UpdateFrameMatrices(CType(frame.FrameSibling, frameD), parentMatrix)
        End If
        'carica i frame figli se ci sono
        If Not (frame.FrameFirstChild Is Nothing) Then
            UpdateFrameMatrices(CType(frame.FrameFirstChild, frameD), frame.matrixC)
        End If
    End Sub

    Public Sub drawMesh()
        'disegna
        DrawFrame(animazione.FrameHierarchy)
    End Sub

    Private Sub DrawFrame(ByRef frame As frameD)

        'disegna il frame
        Dim mesh As MeshContainerDerived = frame.MeshContainer
        'disegna tutti i frame del contenitore
        While Not (mesh Is Nothing)
            DrawMeshContainer(mesh, frame)
            mesh = mesh.NextContainer
        End While
        'disegna i frame nel contenitore se ci sono
        If Not (frame.FrameSibling Is Nothing) Then
            DrawFrame(frame.FrameSibling)
        End If
        'disegna i frame figli se ci sono
        If Not (frame.FrameFirstChild Is Nothing) Then
            DrawFrame(frame.FrameFirstChild)
        End If
    End Sub

    Protected Overridable Sub DrawMeshContainer(ByRef mesh As MeshContainerDerived, ByRef frame As frameD)
        Dim i As Integer
        ' se ci sono skin information disegna la mesh animata in skinning
        If Not (mesh.SkinInformation Is Nothing) Then
            'attribute settato a -1 come non esistente
            Dim attribIdPrev As Integer = -1
            Dim iattrib As Integer
            For iattrib = 0 To mesh.numAttr - 1
                'per ogni attributo nell'oggetto
                Dim numBlend As Integer = 0
                Dim bones() As BoneCombination = mesh.bones 'ricevi le bones

                'numero di blending
                For i = 0 To mesh.numInfl - 1
                    If (bones(iattrib).BoneId(i) <> -1) Then
                        numBlend = i
                    End If
                Next

                If (device.DeviceCaps.MaxVertexBlendMatrices >= numBlend + 1) Then
                    'numero di matrici di blending necessarie
                    device.RenderState.VertexBlend = numBlend
                    'carico le matrici di offset
                    Dim offsetMatrices As Matrix() = mesh.offsetMatrices
                    Dim frameMatrices() As frameD = mesh.frameMatrices
                    'per ogni influenza inserisci la mesh
                    For i = 0 To mesh.numInfl - 1
                        Dim matrixIndex As Integer
                        matrixIndex = bones(iattrib).BoneId(i)
                        If (matrixIndex <> -1) Then
                            Dim tempMatrix As Matrix = Matrix.Multiply(offsetMatrices(matrixIndex), frameMatrices(matrixIndex).matrixC)
                            device.Transform.SetWorldMatrixByIndex(i, tempMatrix)
                        End If
                    Next

                    'inserisci i materiali
                    If ((attribIdPrev <> bones(iattrib).AttributeId) Or attribIdPrev = -1) Then
                        device.Material = mesh.GetMaterials()(bones(iattrib).AttributeId).Material3D
                        device.SetTexture(0, mesh.meshTextures(bones(iattrib).AttributeId))
                        attribIdPrev = bones(iattrib).AttributeId
                    End If
                    'disegna la mesh
                    mesh.MeshData.Mesh.DrawSubset(iattrib)
                End If
            Next

        Else ' normale mesh animata in animation set
            'imposta la matrice
            device.Transform.World = frame.matrixC
            'imposta i materiali
            Dim mtrl() As ExtendedMaterial = mesh.GetMaterials
            Dim iMaterial As Integer
            For iMaterial = 0 To mtrl.Length - 1
                device.Material = mtrl(iMaterial).Material3D
                device.SetTexture(0, mesh.meshTextures(iMaterial))
                'disegna la mesh
                mesh.MeshData.Mesh.DrawSubset(iMaterial)
            Next
        End If
    End Sub

    Private Shared Sub GenerateSkinnedMesh(ByRef mesh As MeshContainerDerived)

        'se non ci sono informazioni esci
        If (mesh.SkinInformation Is Nothing) Then Return
        Dim numInfl As Integer = 0
        Dim bones() As BoneCombination

        ' converti la mesh in blended mesh
        Dim m As MeshData = mesh.MeshData
        m.Mesh = mesh.SkinInformation.ConvertToBlendedMesh(m.Mesh, MeshFlags.Managed, mesh.GetAdjacencyStream, numInfl, bones)
        'salva le informazioni
        mesh.numInfl = numInfl
        mesh.bones = bones
        ' ricevi numero attributi
        mesh.numAttr = bones.Length
        'salva
        mesh.MeshData = m
    End Sub


    'extra informazioni utili
    Public Function getAnimationLenght() As Double
        Return animazione.AnimationController.GetTrackAnimationSet(0).Period
    End Function

    Public Sub resetTime()
        animazione.AnimationController.ResetTime()
        currentTime = 0
    End Sub

    Sub setSpeed(ByVal t As Single)
        animazione.AnimationController.SetTrackSpeed(0, t)
    End Sub

    Public Function getTime() As Double
        Return animazione.AnimationController.Time
    End Function

#Region "Animazioni Multiple"
    Public Sub storeAnimationSets()
        ReDim animationSet(animazione.AnimationController.NumberAnimationSets)
        Dim i As Integer
        For i = 0 To animazione.AnimationController.NumberAnimationSets - 1
            Me.animationSet(i) = animazione.AnimationController.GetAnimationSet(i)
        Next
    End Sub

    Protected Sub BlendAnimation(ByVal goalAnimation As Integer, ByVal resetTimer As Boolean)
        Dim animazioneCorrente As AnimationSet
        Dim prossimaAnimazione As AnimationSet
        Dim newtrack As Byte
        If currentTrack = 0 Then newtrack = 1 Else newtrack = 0
        prossimaAnimazione = animationSet(goalAnimation)

        With animazione.AnimationController
            .SetTrackAnimationSet(newtrack, prossimaAnimazione)
            prossimaAnimazione.Dispose()

            .UnkeyAllTrackEvents(currentTrack)
            .UnkeyAllTrackEvents(newtrack)

            .KeyTrackEnable(currentTrack, False, currentTime + blendTime)
            .KeyTrackSpeed(currentTrack, 0.0F, currentTime, blendTime, TransitionType.Linear)
            .KeyTrackWeight(currentTrack, 0.0F, currentTime, blendTime, TransitionType.Linear)
            .SetTrackEnable(newtrack, True)
            .KeyTrackSpeed(newtrack, 1.0F, currentTime, blendTime, TransitionType.Linear)
            .KeyTrackWeight(newtrack, 1.0F, currentTime, blendTime, TransitionType.Linear)
        End With

        currentTrack = newtrack

        If resetTimer Then
            animazione.AnimationController.ResetTime()
            currentTime = 0
        End If

    End Sub

    Protected Sub setAnimation(ByVal animationSetIndex As Byte)
        Dim anim As AnimationSet
        anim = animationSet(animationSetIndex)
        animazione.AnimationController.SetTrackAnimationSet(0, anim)
        animazione.AnimationController.ResetTime()
        currentTrack = 0
    End Sub
#End Region

End Class
#End Region
